@@ -1,15 +1,16 @@
-# -*- coding: utf-8 -*-
 """
     babel.messages.frontend
     ~~~~~~~~~~~~~~~~~~~~~~~
 
     Frontends for the message extraction functionality.
 
-    :copyright: (c) 2013-2022 by the Babel Team.
+    :copyright: (c) 2013-2024 by the Babel Team.
     :license: BSD, see LICENSE for more details.
 """
-from __future__ import print_function
 
+from __future__ import annotations
+
+import datetime
 import fnmatch
 import logging
 import optparse
@@ -20,35 +21,36 @@ import sys
 import tempfile
 from collections import OrderedDict
 from configparser import RawConfigParser
-from datetime import datetime
 from io import StringIO
+from typing import Iterable
 
-from babel import __version__ as VERSION
 from babel import Locale, localedata
+from babel import __version__ as VERSION
 from babel.core import UnknownLocaleError
-from babel.messages.catalog import Catalog, DEFAULT_HEADER
-from babel.messages.extract import DEFAULT_KEYWORDS, DEFAULT_MAPPING, check_and_call_extract_file, extract_from_dir
+from babel.messages.catalog import DEFAULT_HEADER, Catalog
+from babel.messages.extract import (
+    DEFAULT_KEYWORDS,
+    DEFAULT_MAPPING,
+    check_and_call_extract_file,
+    extract_from_dir,
+)
 from babel.messages.mofile import write_mo
 from babel.messages.pofile import read_po, write_po
 from babel.util import LOCALTZ
 
 log = logging.getLogger('babel')
 
-try:
-    # See: https://setuptools.pypa.io/en/latest/deprecated/distutils-legacy.html
-    from setuptools import Command as _Command
-    distutils_log = log  # "distutils.log → (no replacement yet)"
 
-    try:
-        from setuptools.errors import OptionError, SetupError, BaseError
-    except ImportError:  # Error aliases only added in setuptools 59 (2021-11).
-        OptionError = SetupError = BaseError = Exception
+class BaseError(Exception):
+    pass
 
-except ImportError:
-    from distutils import log as distutils_log
-    from distutils.cmd import Command as _Command
-    from distutils.errors import DistutilsOptionError as OptionError, DistutilsSetupError as SetupError, DistutilsError as BaseError
 
+class OptionError(BaseError):
+    pass
+
+
+class SetupError(BaseError):
+    pass
 
 
 def listify_value(arg, split=None):
@@ -93,7 +95,7 @@ def listify_value(arg, split=None):
     return out
 
 
-class Command(_Command):
+class CommandMixin:
     # This class is a small shim between Distutils commands and
     # optparse option parsing in the frontend command line.
 
@@ -121,7 +123,7 @@ class Command(_Command):
     option_choices = {}
 
     #: Log object. To allow replacement in the script command line runner.
-    log = distutils_log
+    log = log
 
     def __init__(self, dist=None):
         # A less strict version of distutils' `__init__`.
@@ -133,24 +135,21 @@ class Command(_Command):
         self.help = 0
         self.finalized = 0
 
+    def initialize_options(self):
+        pass
 
-class compile_catalog(Command):
-    """Catalog compilation command for use in ``setup.py`` scripts.
+    def ensure_finalized(self):
+        if not self.finalized:
+            self.finalize_options()
+        self.finalized = 1
 
-    If correctly installed, this command is available to Setuptools-using
-    setup scripts automatically. For projects using plain old ``distutils``,
-    the command needs to be registered explicitly in ``setup.py``::
-
-        from babel.messages.frontend import compile_catalog
-
-        setup(
-            ...
-            cmdclass = {'compile_catalog': compile_catalog}
+    def finalize_options(self):
+        raise RuntimeError(
+            f"abstract method -- subclass {self.__class__} must override",
         )
 
-    .. versionadded:: 0.9
-    """
 
+class CompileCatalog(CommandMixin):
     description = 'compile message catalogs to binary MO files'
     user_options = [
         ('domain=', 'D',
@@ -167,7 +166,7 @@ class compile_catalog(Command):
         ('use-fuzzy', 'f',
          'also include fuzzy translations'),
         ('statistics', None,
-         'print statistics about translations')
+         'print statistics about translations'),
     ]
     boolean_options = ['use-fuzzy', 'statistics']
 
@@ -190,10 +189,10 @@ class compile_catalog(Command):
     def run(self):
         n_errors = 0
         for domain in self.domain:
-            for catalog, errors in self._run_domain(domain).items():
+            for errors in self._run_domain(domain).values():
                 n_errors += len(errors)
         if n_errors:
-            self.log.error('%d errors encountered.' % n_errors)
+            self.log.error('%d errors encountered.', n_errors)
         return (1 if n_errors else 0)
 
     def _run_domain(self, domain):
@@ -205,19 +204,19 @@ class compile_catalog(Command):
                 po_files.append((self.locale,
                                  os.path.join(self.directory, self.locale,
                                               'LC_MESSAGES',
-                                              domain + '.po')))
+                                              f"{domain}.po")))
                 mo_files.append(os.path.join(self.directory, self.locale,
                                              'LC_MESSAGES',
-                                             domain + '.mo'))
+                                             f"{domain}.mo"))
             else:
                 for locale in os.listdir(self.directory):
                     po_file = os.path.join(self.directory, locale,
-                                           'LC_MESSAGES', domain + '.po')
+                                           'LC_MESSAGES', f"{domain}.po")
                     if os.path.exists(po_file):
                         po_files.append((locale, po_file))
                         mo_files.append(os.path.join(self.directory, locale,
                                                      'LC_MESSAGES',
-                                                     domain + '.mo'))
+                                                     f"{domain}.mo"))
         else:
             po_files.append((self.locale, self.input_file))
             if self.output_file:
@@ -225,7 +224,7 @@ class compile_catalog(Command):
             else:
                 mo_files.append(os.path.join(self.directory, self.locale,
                                              'LC_MESSAGES',
-                                             domain + '.mo'))
+                                             f"{domain}.mo"))
 
         if not po_files:
             raise OptionError('no message catalogs found')
@@ -247,7 +246,7 @@ class compile_catalog(Command):
                     percentage = translated * 100 // len(catalog)
                 self.log.info(
                     '%d of %d messages (%d%%) translated in %s',
-                    translated, len(catalog), percentage, po_file
+                    translated, len(catalog), percentage, po_file,
                 )
 
             if catalog.fuzzy and not self.use_fuzzy:
@@ -258,7 +257,7 @@ class compile_catalog(Command):
             for message, errors in catalog_errors:
                 for error in errors:
                     self.log.error(
-                        'error: %s:%d: %s', po_file, message.lineno, error
+                        'error: %s:%d: %s', po_file, message.lineno, error,
                     )
 
             self.log.info('compiling catalog %s to %s', po_file, mo_file)
@@ -273,6 +272,7 @@ def _make_directory_filter(ignore_patterns):
     """
     Build a directory_filter function based on a list of ignore patterns.
     """
+
     def cli_directory_filter(dirname):
         basename = os.path.basename(dirname)
         return not any(
@@ -280,24 +280,11 @@ def _make_directory_filter(ignore_patterns):
             for ignore_pattern
             in ignore_patterns
         )
+
     return cli_directory_filter
 
 
-class extract_messages(Command):
-    """Message extraction command for use in ``setup.py`` scripts.
-
-    If correctly installed, this command is available to Setuptools-using
-    setup scripts automatically. For projects using plain old ``distutils``,
-    the command needs to be registered explicitly in ``setup.py``::
-
-        from babel.messages.frontend import extract_messages
-
-        setup(
-            ...
-            cmdclass = {'extract_messages': extract_messages}
-        )
-    """
-
+class ExtractMessages(CommandMixin):
     description = 'extract localizable strings from the project code'
     user_options = [
         ('charset=', None,
@@ -352,10 +339,12 @@ class extract_messages(Command):
          'Separate multiple patterns with spaces (default ".* ._")'),
         ('header-comment=', None,
          'header comment for the catalog'),
+        ('last-translator=', None,
+         'set the name and email of the last translator in output'),
     ]
     boolean_options = [
         'no-default-keywords', 'no-location', 'omit-header', 'no-wrap',
-        'sort-output', 'sort-by-file', 'strip-comments'
+        'sort-output', 'sort-by-file', 'strip-comments',
     ]
     as_args = 'input-paths'
     multiple_value_options = (
@@ -368,9 +357,10 @@ class extract_messages(Command):
         'mapping-file': ('--mapping',),
         'output-file': ('--output',),
         'strip-comments': ('--strip-comment-tags',),
+        'last-translator': ('--last-translator',),
     }
     option_choices = {
-        'add-location': ('full', 'file', 'never',),
+        'add-location': ('full', 'file', 'never'),
     }
 
     def initialize_options(self):
@@ -397,6 +387,7 @@ class extract_messages(Command):
         self.include_lineno = True
         self.ignore_dirs = None
         self.header_comment = None
+        self.last_translator = None
 
     def finalize_options(self):
         if self.input_dirs:
@@ -404,13 +395,10 @@ class extract_messages(Command):
                 self.input_paths = self.input_dirs
             else:
                 raise OptionError(
-                    'input-dirs and input-paths are mutually exclusive'
+                    'input-dirs and input-paths are mutually exclusive',
                 )
 
-        if self.no_default_keywords:
-            keywords = {}
-        else:
-            keywords = DEFAULT_KEYWORDS.copy()
+        keywords = {} if self.no_default_keywords else DEFAULT_KEYWORDS.copy()
 
         keywords.update(parse_keywords(listify_value(self.keywords)))
 
@@ -418,14 +406,14 @@ class extract_messages(Command):
 
         if not self.keywords:
             raise OptionError(
-                'you must specify new keywords if you disable the default ones'
+                'you must specify new keywords if you disable the default ones',
             )
 
         if not self.output_file:
             raise OptionError('no output file specified')
         if self.no_wrap and self.width:
             raise OptionError(
-                "'--no-wrap' and '--width' are mutually exclusive"
+                "'--no-wrap' and '--width' are mutually exclusive",
             )
         if not self.no_wrap and not self.width:
             self.width = 76
@@ -434,7 +422,7 @@ class extract_messages(Command):
 
         if self.sort_output and self.sort_by_file:
             raise OptionError(
-                "'--sort-output' and '--sort-by-file' are mutually exclusive"
+                "'--sort-output' and '--sort-by-file' are mutually exclusive",
             )
 
         if self.input_paths:
@@ -453,7 +441,7 @@ class extract_messages(Command):
 
         for path in self.input_paths:
             if not os.path.exists(path):
-                raise OptionError("Input path: %s does not exist" % path)
+                raise OptionError(f"Input path: {path} does not exist")
 
         self.add_comments = listify_value(self.add_comments or (), ",")
 
@@ -474,6 +462,28 @@ class extract_messages(Command):
         else:
             self.directory_filter = None
 
+    def _build_callback(self, path: str):
+        def callback(filename: str, method: str, options: dict):
+            if method == 'ignore':
+                return
+
+            # If we explicitly provide a full filepath, just use that.
+            # Otherwise, path will be the directory path and filename
+            # is the relative path from that dir to the file.
+            # So we can join those to get the full filepath.
+            if os.path.isfile(path):
+                filepath = path
+            else:
+                filepath = os.path.normpath(os.path.join(path, filename))
+
+            optstr = ''
+            if options:
+                opt_values = ", ".join(f'{k}="{v}"' for k, v in options.items())
+                optstr = f" ({opt_values})"
+            self.log.info('extracting messages from %s%s', filepath, optstr)
+
+        return callback
+
     def run(self):
         mappings = self._get_mappings()
         with open(self.output_file, 'wb') as outfile:
@@ -482,34 +492,17 @@ class extract_messages(Command):
                               msgid_bugs_address=self.msgid_bugs_address,
                               copyright_holder=self.copyright_holder,
                               charset=self.charset,
-                              header_comment=(self.header_comment or DEFAULT_HEADER))
+                              header_comment=(self.header_comment or DEFAULT_HEADER),
+                              last_translator=self.last_translator)
 
             for path, method_map, options_map in mappings:
-                def callback(filename, method, options):
-                    if method == 'ignore':
-                        return
-
-                    # If we explicitly provide a full filepath, just use that.
-                    # Otherwise, path will be the directory path and filename
-                    # is the relative path from that dir to the file.
-                    # So we can join those to get the full filepath.
-                    if os.path.isfile(path):
-                        filepath = path
-                    else:
-                        filepath = os.path.normpath(os.path.join(path, filename))
-
-                    optstr = ''
-                    if options:
-                        optstr = ' (%s)' % ', '.join(['%s="%s"' % (k, v) for
-                                                      k, v in options.items()])
-                    self.log.info('extracting messages from %s%s', filepath, optstr)
-
+                callback = self._build_callback(path)
                 if os.path.isfile(path):
                     current_dir = os.getcwd()
                     extracted = check_and_call_extract_file(
                         path, method_map, options_map,
                         callback, self.keywords, self.add_comments,
-                        self.strip_comments, current_dir
+                        self.strip_comments, current_dir,
                     )
                 else:
                     extracted = extract_from_dir(
@@ -565,38 +558,7 @@ class extract_messages(Command):
         return mappings
 
 
-def check_message_extractors(dist, name, value):
-    """Validate the ``message_extractors`` keyword argument to ``setup()``.
-
-    :param dist: the distutils/setuptools ``Distribution`` object
-    :param name: the name of the keyword argument (should always be
-                 "message_extractors")
-    :param value: the value of the keyword argument
-    :raise `DistutilsSetupError`: if the value is not valid
-    """
-    assert name == 'message_extractors'
-    if not isinstance(value, dict):
-        raise SetupError(
-            'the value of the "message_extractors" '
-            'parameter must be a dictionary'
-        )
-
-
-class init_catalog(Command):
-    """New catalog initialization command for use in ``setup.py`` scripts.
-
-    If correctly installed, this command is available to Setuptools-using
-    setup scripts automatically. For projects using plain old ``distutils``,
-    the command needs to be registered explicitly in ``setup.py``::
-
-        from babel.messages.frontend import init_catalog
-
-        setup(
-            ...
-            cmdclass = {'init_catalog': init_catalog}
-        )
-    """
-
+class InitCatalog(CommandMixin):
     description = 'create a new catalog based on a POT file'
     user_options = [
         ('domain=', 'D',
@@ -642,7 +604,7 @@ class init_catalog(Command):
             raise OptionError('you must specify the output directory')
         if not self.output_file:
             self.output_file = os.path.join(self.output_dir, self.locale,
-                                            'LC_MESSAGES', self.domain + '.po')
+                                            'LC_MESSAGES', f"{self.domain}.po")
 
         if not os.path.exists(os.path.dirname(self.output_file)):
             os.makedirs(os.path.dirname(self.output_file))
@@ -655,7 +617,7 @@ class init_catalog(Command):
 
     def run(self):
         self.log.info(
-            'creating catalog %s based on %s', self.output_file, self.input_file
+            'creating catalog %s based on %s', self.output_file, self.input_file,
         )
 
         with open(self.input_file, 'rb') as infile:
@@ -664,30 +626,14 @@ class init_catalog(Command):
             catalog = read_po(infile, locale=self.locale)
 
         catalog.locale = self._locale
-        catalog.revision_date = datetime.now(LOCALTZ)
+        catalog.revision_date = datetime.datetime.now(LOCALTZ)
         catalog.fuzzy = False
 
         with open(self.output_file, 'wb') as outfile:
             write_po(outfile, catalog, width=self.width)
 
 
-class update_catalog(Command):
-    """Catalog merging command for use in ``setup.py`` scripts.
-
-    If correctly installed, this command is available to Setuptools-using
-    setup scripts automatically. For projects using plain old ``distutils``,
-    the command needs to be registered explicitly in ``setup.py``::
-
-        from babel.messages.frontend import update_catalog
-
-        setup(
-            ...
-            cmdclass = {'update_catalog': update_catalog}
-        )
-
-    .. versionadded:: 0.9
-    """
-
+class UpdateCatalog(CommandMixin):
     description = 'update message catalogs from a POT file'
     user_options = [
         ('domain=', 'D',
@@ -722,11 +668,13 @@ class update_catalog(Command):
          'don\'t update the catalog, just return the status. Return code 0 '
          'means nothing would change. Return code 1 means that the catalog '
          'would be updated'),
+        ('ignore-pot-creation-date=', None,
+         'ignore changes to POT-Creation-Date when updating or checking'),
     ]
     boolean_options = [
         'omit-header', 'no-wrap', 'ignore-obsolete', 'init-missing',
         'no-fuzzy-matching', 'previous', 'update-header-comment',
-        'check',
+        'check', 'ignore-pot-creation-date',
     ]
 
     def initialize_options(self):
@@ -744,6 +692,7 @@ class update_catalog(Command):
         self.update_header_comment = False
         self.previous = False
         self.check = False
+        self.ignore_pot_creation_date = False
 
     def finalize_options(self):
         if not self.input_file:
@@ -757,7 +706,7 @@ class update_catalog(Command):
             if not self.locale:
                 raise OptionError(
                     'you must specify the locale for '
-                    'the init-missing option to work'
+                    'the init-missing option to work',
                 )
 
             try:
@@ -784,12 +733,12 @@ class update_catalog(Command):
                 po_files.append((self.locale,
                                  os.path.join(self.output_dir, self.locale,
                                               'LC_MESSAGES',
-                                              self.domain + '.po')))
+                                              f"{self.domain}.po")))
             else:
                 for locale in os.listdir(self.output_dir):
                     po_file = os.path.join(self.output_dir, locale,
                                            'LC_MESSAGES',
-                                           self.domain + '.po')
+                                           f"{self.domain}.po")
                     if os.path.exists(po_file):
                         po_files.append((locale, po_file))
         else:
@@ -811,7 +760,7 @@ class update_catalog(Command):
                     check_status[filename] = False
                     continue
                 self.log.info(
-                    'creating catalog %s based on %s', filename, self.input_file
+                    'creating catalog %s based on %s', filename, self.input_file,
                 )
 
                 with open(self.input_file, 'rb') as infile:
@@ -820,7 +769,7 @@ class update_catalog(Command):
                     catalog = read_po(infile, locale=self.locale)
 
                 catalog.locale = self._locale
-                catalog.revision_date = datetime.now(LOCALTZ)
+                catalog.revision_date = datetime.datetime.now(LOCALTZ)
                 catalog.fuzzy = False
 
                 with open(filename, 'wb') as outfile:
@@ -832,7 +781,8 @@ class update_catalog(Command):
 
             catalog.update(
                 template, self.no_fuzzy_matching,
-                update_header_comment=self.update_header_comment
+                update_header_comment=self.update_header_comment,
+                update_creation_date=not self.ignore_pot_creation_date,
             )
 
             tmpname = os.path.join(os.path.dirname(filename),
@@ -844,7 +794,7 @@ class update_catalog(Command):
                              omit_header=self.omit_header,
                              ignore_obsolete=self.ignore_obsolete,
                              include_previous=self.previous, width=self.width)
-            except:
+            except Exception:
                 os.remove(tmpname)
                 raise
 
@@ -883,7 +833,7 @@ class update_catalog(Command):
             return
 
 
-class CommandLineInterface(object):
+class CommandLineInterface:
     """Command-line interface.
 
     This class provides a simple command-line interface to the message
@@ -891,19 +841,19 @@ class CommandLineInterface(object):
     """
 
     usage = '%%prog %s [options] %s'
-    version = '%%prog %s' % VERSION
+    version = f'%prog {VERSION}'
     commands = {
         'compile': 'compile message catalogs to MO files',
         'extract': 'extract messages from source files and generate a POT file',
         'init': 'create new message catalogs from a POT file',
-        'update': 'update existing message catalogs from a POT file'
+        'update': 'update existing message catalogs from a POT file',
     }
 
     command_classes = {
-        'compile': compile_catalog,
-        'extract': extract_messages,
-        'init': init_catalog,
-        'update': update_catalog,
+        'compile': CompileCatalog,
+        'extract': ExtractMessages,
+        'init': InitCatalog,
+        'update': UpdateCatalog,
     }
 
     log = None  # Replaced on instance level
@@ -937,12 +887,10 @@ class CommandLineInterface(object):
         self._configure_logging(options.loglevel)
         if options.list_locales:
             identifiers = localedata.locale_identifiers()
-            longest = max([len(identifier) for identifier in identifiers])
-            identifiers.sort()
-            format = u'%%-%ds %%s' % (longest + 1)
-            for identifier in identifiers:
+            id_width = max(len(identifier) for identifier in identifiers) + 1
+            for identifier in sorted(identifiers):
                 locale = Locale.parse(identifier)
-                print(format % (identifier, locale.english_name))
+                print(f"{identifier:<{id_width}} {locale.english_name}")
             return 0
 
         if not args:
@@ -951,7 +899,7 @@ class CommandLineInterface(object):
 
         cmdname = args[0]
         if cmdname not in self.commands:
-            self.parser.error('unknown command "%s"' % cmdname)
+            self.parser.error(f'unknown command "{cmdname}"')
 
         cmdinst = self._configure_command(cmdname, args[1:])
         return cmdinst.run()
@@ -974,11 +922,9 @@ class CommandLineInterface(object):
     def _help(self):
         print(self.parser.format_help())
         print("commands:")
-        longest = max([len(command) for command in self.commands])
-        format = "  %%-%ds %%s" % max(8, longest + 1)
-        commands = sorted(self.commands.items())
-        for name, description in commands:
-            print(format % (name, description))
+        cmd_width = max(8, max(len(command) for command in self.commands) + 1)
+        for name, description in sorted(self.commands.items()):
+            print(f"  {name:<{cmd_width}} {description}")
 
     def _configure_command(self, cmdname, argv):
         """
@@ -989,24 +935,24 @@ class CommandLineInterface(object):
         cmdinst = cmdclass()
         if self.log:
             cmdinst.log = self.log  # Use our logger, not distutils'.
-        assert isinstance(cmdinst, Command)
+        assert isinstance(cmdinst, CommandMixin)
         cmdinst.initialize_options()
 
         parser = optparse.OptionParser(
             usage=self.usage % (cmdname, ''),
-            description=self.commands[cmdname]
+            description=self.commands[cmdname],
         )
         as_args = getattr(cmdclass, "as_args", ())
         for long, short, help in cmdclass.user_options:
             name = long.strip("=")
-            default = getattr(cmdinst, name.replace('-', '_'))
-            strs = ["--%s" % name]
+            default = getattr(cmdinst, name.replace("-", "_"))
+            strs = [f"--{name}"]
             if short:
-                strs.append("-%s" % short)
+                strs.append(f"-{short}")
             strs.extend(cmdclass.option_aliases.get(name, ()))
             choices = cmdclass.option_choices.get(name, None)
             if name == as_args:
-                parser.usage += "<%s>" % name
+                parser.usage += f"<{name}>"
             elif name in cmdclass.boolean_options:
                 parser.add_option(*strs, action="store_true", help=help)
             elif name in cmdclass.multiple_value_options:
@@ -1094,7 +1040,7 @@ def parse_mapping(fileobj, filename=None):
         if section == 'extractors':
             extractors = dict(parser.items(section))
         else:
-            method, pattern = [part.strip() for part in section.split(':', 1)]
+            method, pattern = (part.strip() for part in section.split(':', 1))
             method_map.append((pattern, method))
             options_map[pattern] = dict(parser.items(section))
 
@@ -1107,35 +1053,77 @@ def parse_mapping(fileobj, filename=None):
     return method_map, options_map
 
 
-def parse_keywords(strings=[]):
+def _parse_spec(s: str) -> tuple[int | None, tuple[int | tuple[int, str], ...]]:
+    inds = []
+    number = None
+    for x in s.split(','):
+        if x[-1] == 't':
+            number = int(x[:-1])
+        elif x[-1] == 'c':
+            inds.append((int(x[:-1]), 'c'))
+        else:
+            inds.append(int(x))
+    return number, tuple(inds)
+
+
+def parse_keywords(strings: Iterable[str] = ()):
     """Parse keywords specifications from the given list of strings.
 
-    >>> kw = sorted(parse_keywords(['_', 'dgettext:2', 'dngettext:2,3', 'pgettext:1c,2']).items())
-    >>> for keyword, indices in kw:
-    ...     print((keyword, indices))
-    ('_', None)
-    ('dgettext', (2,))
-    ('dngettext', (2, 3))
-    ('pgettext', ((1, 'c'), 2))
+    >>> import pprint
+    >>> keywords = ['_', 'dgettext:2', 'dngettext:2,3', 'pgettext:1c,2',
+    ...             'polymorphic:1', 'polymorphic:2,2t', 'polymorphic:3c,3t']
+    >>> pprint.pprint(parse_keywords(keywords))
+    {'_': None,
+     'dgettext': (2,),
+     'dngettext': (2, 3),
+     'pgettext': ((1, 'c'), 2),
+     'polymorphic': {None: (1,), 2: (2,), 3: ((3, 'c'),)}}
+
+    The input keywords are in GNU Gettext style; see :doc:`cmdline` for details.
+
+    The output is a dictionary mapping keyword names to a dictionary of specifications.
+    Keys in this dictionary are numbers of arguments, where ``None`` means that all numbers
+    of arguments are matched, and a number means only calls with that number of arguments
+    are matched (which happens when using the "t" specifier). However, as a special
+    case for backwards compatibility, if the dictionary of specifications would
+    be ``{None: x}``, i.e., there is only one specification and it matches all argument
+    counts, then it is collapsed into just ``x``.
+
+    A specification is either a tuple or None. If a tuple, each element can be either a number
+    ``n``, meaning that the nth argument should be extracted as a message, or the tuple
+    ``(n, 'c')``, meaning that the nth argument should be extracted as context for the
+    messages. A ``None`` specification is equivalent to ``(1,)``, extracting the first
+    argument.
     """
     keywords = {}
     for string in strings:
         if ':' in string:
-            funcname, indices = string.split(':')
+            funcname, spec_str = string.split(':')
+            number, spec = _parse_spec(spec_str)
         else:
-            funcname, indices = string, None
-        if funcname not in keywords:
-            if indices:
-                inds = []
-                for x in indices.split(','):
-                    if x[-1] == 'c':
-                        inds.append((int(x[:-1]), 'c'))
-                    else:
-                        inds.append(int(x))
-                indices = tuple(inds)
-            keywords[funcname] = indices
+            funcname = string
+            number = None
+            spec = None
+        keywords.setdefault(funcname, {})[number] = spec
+
+    # For best backwards compatibility, collapse {None: x} into x.
+    for k, v in keywords.items():
+        if set(v) == {None}:
+            keywords[k] = v[None]
+
     return keywords
 
 
+def __getattr__(name: str):
+    # Re-exports for backwards compatibility;
+    # `setuptools_frontend` is the canonical import location.
+    if name in {'check_message_extractors', 'compile_catalog', 'extract_messages', 'init_catalog', 'update_catalog'}:
+        from babel.messages import setuptools_frontend
+
+        return getattr(setuptools_frontend, name)
+
+    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")
+
+
 if __name__ == '__main__':
     main()
